@@ -1,5 +1,3 @@
-// Copyright (c) 2020 Red Hat, Inc.
-
 /*
 Copyright 2015 The Kubernetes Authors.
 
@@ -16,6 +14,9 @@ See the License for the specific language governing permissions and
 limitations under the License.
 */
 
+// Copyright (c) 2021 Red Hat, Inc.
+// Copyright Contributors to the Open Cluster Management project
+
 package ssl
 
 import (
@@ -41,8 +42,8 @@ import (
 
 	"k8s.io/apimachinery/pkg/util/sets"
 
-	"github.com/open-cluster-management/management-ingress/pkg/file"
-	"github.com/open-cluster-management/management-ingress/pkg/ingress"
+	"github.com/stolostron/management-ingress/pkg/file"
+	"github.com/stolostron/management-ingress/pkg/ingress"
 )
 
 var (
